/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 University of Pennsylvania
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ol-context.h"
#include "ol-lexer.h"

using namespace ns3;
using namespace ns3::rapidnet_compiler;

NS_LOG_COMPONENT_DEFINE ("OlContext");

/** No expiration is represented as positive infinity. */
long OlContext::NO_EXPIRATION = -1L;

/** Default table expiration time is no expiration */
long OlContext::DEFAULT_EXPIRATION = OlContext::NO_EXPIRATION;

/** No size is represented as 0 size. */
uint32_t OlContext::NO_SIZE = 0;

/** Default table max size is no max size */
uint32_t OlContext::DEFAULT_SIZE (OlContext::NO_SIZE);

int OlContext::ruleCount = 0;

/**********************************************************************
 *
 * Rule methods
 *
 *********************************************************************/

//
// Print out the rule for debugging purposes
//
string OlContext::Rule::ToString ()
{
  ostringstream r;
  r << ruleID << " ";
  r << head->ToString () << " :- ";

  list<ParseTerm*>::iterator t = terms.begin ();
  r << (*t)->ToString ();
  t++;
  for (; t != terms.end (); t++)
    {
      r << ", " << (*t)->ToString ();
    }
  r << ".";

  return r.str ();
}

string OlContext::TableInfo::ToString ()
{
  ostringstream t;
  t << "materialize(" << tableName << ", " << timeout << ", " << size
      << ", keys(";

  for (uint k = 0; k + 1 < primaryKeys.size (); k++)
    {
      t << primaryKeys.at (k) << ", ";
    }
  if (primaryKeys.size () > 0)
    {
      t << primaryKeys.at (primaryKeys.size () - 1);
    }
  t << ")";

  return t.str ();
}

string OlContext::Context::ToString ()
{
  return mId->ToString ();
}

void OlContext::SetContext (ParseExpr *ctx)
{
  mContext.mId = ctx;
  dynamic_cast<ParseVar*> (mContext.mId) ->SetLocSpec ();
  //cout<<"set context: "<<loc_context.ToString()<<endl;
}

/**********************************************************************
 *
 * Context methods
 *
 *********************************************************************/

//
// Adding a new rule to the environment.  Subtle: we need to resolve
// all variable references at this point and convert them to
// positional arguments.
//
OlContext::Rule* OlContext::CreateRule (ParseTerm *lhs, ParseTermList *rhs,
  bool deleteFlag, ParseExpr *n, bool deleteEventFlag)
{
  //  TRACE_FUNCTION;

  //SeNDlog: Rewrite flag is set when context is set.
  bool sendlog_rewrite = IsSendlog ();

  ParseFunctor *h = dynamic_cast<ParseFunctor*> (lhs);
  string ruleName = (n) ? n->value->ToString () : "";
  int fict_varnum = 1; // Counter for inventing anonymous variables.


  // Get hold of the functor (rule head) for this new rule.  This
  // holds a list of all the rule bodies that match the head.  Note
  // that this match is only performed on (name,arity), since we
  // convert all bound values to extra "eq" terms in the body down
  // below.

  // Create a new rule and register it.
  Rule *r = new Rule (ruleName, h, deleteFlag, deleteEventFlag);

  // Next, we canonicalize all the args in the rule head.  We build up
  // a list of argument names - the first 'arity' of these will be the
  // free variables in the rule head.  Literals and duplicate free
  // variables here are eliminated, by a process of appending extra
  // "eq" terms to the body, and inventing new free variables.

  //SeNDlog: Outgoing flag is set when a header tuple is going out, in this case
  //the destination need to be taken into the functor and set as location
  //specifier.
  bool outgoing = false;
  if (sendlog_rewrite)
    {
      outgoing = true;
    }

  //SeNDlog: The sayer of an authenticated tuple needs to be considered when doing join.
  //So take sayer into the tuple if it is not already there.
  //To check whether the sayer is already there, assure not first.
  bool has_sayer = true;
  if (h->m_auth)
    {
      has_sayer = false;
    }

  //SeNDlog: Syntax check. There should be no @ in the rules.
  if (sendlog_rewrite && (!h->m_loc.empty ()))
    {
      if (sendlog_rewrite)
        {
          ReportError (
            "Location specifier not allowed in SeNDlog rules.");
        }
    }

  for (int i = 0; i < h->Args (); i++)
    {
      ParseAgg *agg = NULL;
      ParseVar *var = NULL;
      ParseVal *val = NULL;

      if ( (agg = dynamic_cast<ParseAgg*> (h->Arg (i))) != NULL)
        {
          agg->Position (i);
        }
      else if ( (var = dynamic_cast<ParseVar*> (h->Arg (i))) != NULL)
        {
          // The argument is a free variable - the usual case.

          //SeNDlog: Do SeNDlog re-write when the flag is set.
          if (sendlog_rewrite)
            {
              //Check if this variable is the sayer when this is an authenticated tuple.
              if (h->m_auth)
                {
                  if (var->ToString () == h->m_sayer->ToString ())
                    {
                      has_sayer = true;
                    }
                }

              //No out going destination, this is a local functor.
              //The location specifier for this functor should be the context.
              if (!h->m_dest)
                {
                  outgoing = false;
                  //The variable which equals the context should be set location specifier.
                  if (var->ToString ()
                      == dynamic_cast<ParseVar*> (mContext.mId)->ToString ())
                    {
                      //If no location specifier yet, set this variable.
                      if (h->m_loc.empty ())
                        {
                          var->SetLocSpec ();
                        }
                    }
                }
              //Has out going destination, this is an outgoing functor.
              //The location specifier for this functor should be the destination.
              else
                {
                  //The variable which equals the destination should be set location specifier.
                  if (var->ToString () == h->m_dest->ToString ())
                    {
                      var->SetLocSpec ();
                      outgoing = false; //Turn off the flag since this functor is located.
                    }
                }
            }

          int loc = h->Find (var->ToString ());
          if (loc < i)
            {
              ostringstream oss;
              oss << "$" << fict_varnum++;
              // We've found a duplicate variable in the head. Add a new
              // "eq" term to the front of the term list.
              ParseVar *tmp = new ParseVar (oss.str ());
              tmp->Position (i);
              h->replace (i, tmp);
              r->terms.push_back (new ParseAssign (tmp, h->Arg (loc)));
            }
          else
            {
              var->Position (i);
            }
        }
      else if ( (val = dynamic_cast<ParseVal*> (h->Arg (i))) != NULL)
        {
          ostringstream oss;
          oss << "$" << fict_varnum++;
          ParseVar *tmp = new ParseVar (oss.str ());
          tmp->Position (i);
          h->replace (i, tmp);
          r->terms.push_back (new ParseAssign (tmp, val));
        }
      else
        {
          ReportError ("Parse rule unknown functor body type.");
        }

    }

  //SeNDlog: Previously the location specifier is only set on in the variable.
  //Now the functor will locate its locations specifier variable.
  if (sendlog_rewrite)
    {
      //For out going functors whose destination dosen't appear in the
      //the variables, take in the destination explicitly and set is as
      //location specifier.
      if (outgoing)
        {
          ParseVar *d = new ParseVar (h->m_dest);
          d->SetLocSpec ();
          h->m_args->push_back (d);
        }
      //Other cases, the location specifier should be set already.
      //Just let the functor to relocate it.
      else
        {
          (void) h->GetLocSpec ();
        }

      //Add the sayer into the tuple if it isn't already there.
      //And always do this last to make sure it will be at the same position
      //as in body functor rewrite.
      if (!has_sayer)
        {
          h->arg (h->m_sayer);
        }

    }

  // Now we've taken care of the head (and possibly created a few
  // terms), we run through the rest of the terms converting all the
  // variables we encounter to indices.
  int tpos = 1;
  for (ParseTermList::iterator iter = rhs->begin (); iter != rhs->end (); (*iter)->Position (
    tpos++), iter++)
    {
      ParseFunctor *f = NULL;
      ParseAssign *a = NULL;
      ParseSelect *s = NULL;

      if ( (f = dynamic_cast<ParseFunctor*> (*iter)) != NULL)
        {
          //SeNDlog: not_located flag is on for rule body tuples before the re-write
          //assign it the location specifier through parse the variables in the tuple.
          //In this case, the context needs to be added into the tuple explicitly as
          //the location specifier.
          bool not_located = false;
          if (sendlog_rewrite)
            {
              not_located = true;
            }

          //SeNDlog: The sayer of an authenticated tuple needs to be considered when doing join.
          //So take sayer into the tuple if it is not already there.
          //To check whether the sayer is already there, assure not first.
          bool has_sayer = true;
          if (f->m_auth)
            {
              has_sayer = false;
            }

          //SeNDlog: Syntax check. There should be no @ in the rules.
          if (sendlog_rewrite && (!f->m_loc.empty ()))
            {
              if (sendlog_rewrite)
                ReportError (
                  "[SeNDlog] There should be no location specifier in the rules.");
            }

          for (int i = 0; i < f->Args (); i++)
            {
              ParseVar *var = NULL;
              ParseVal *val = NULL;

              if ( (var = dynamic_cast<ParseVar*> (f->Arg (i))) != NULL)
                {
                  //SeNDlog: Do SeNDlog re-write when the flag is set.
                  if (sendlog_rewrite)
                    {
                      //check if this variable is the sayer when this is an authenticated tuple.
                      if (f->m_sayer != NULL)
                        {
                          if (var->ToString () == f->m_sayer->ToString ())
                            {
                              has_sayer = true;
                            }
                        }

                      //All functors in the rule body should be local.
                      //The variable which equals the context should be set location specifier.
                      if (var->ToString ()
                          == dynamic_cast<ParseVar*> (mContext.mId)->ToString ())
                        {
                          if (f->m_loc.empty ())
                            {
                              var->SetLocSpec ();
                            }
                          not_located = false; //located, turn of this flag.
                        }
                    }

                  var->Position (i);
                }
              else if ( (val = dynamic_cast<ParseVal*> (f->Arg (i))) != NULL)
                {
                  val->Position (i);
                }
              else
                {
                  ReportError ("Parse functor term unknown argument type.");
                }
            }

          //SeNDlog: Previously the location specifier is only set on in the variable.
          //Now the functor will locate its locations specifier variable.
          if (sendlog_rewrite)
            {
              //For body functors whose context dosen't appear in the
              //the variables, take in the context explicitly and set is as
              //location specifier.
              if (not_located)
                {
                  f->arg (mContext.mId);
                }

              (void) f->GetLocSpec (); //relocate

              //Add the sayer into the tuple if it isn't already there.
              //And always do this last to make sure it will be at the same position
              //as in body functor rewrite.
              if (!has_sayer)
                {
                  f->arg (f->m_sayer);
                }

            }

        }
      else if ( (a = dynamic_cast<ParseAssign*> (*iter)) != NULL)
        {

        }
      else if ( (s = dynamic_cast<ParseSelect*> (*iter)) != NULL)
        {

        }
      else
        {
          ReportError ("Internal parse error: unknown term type!");
        }
      r->terms.push_back (*iter);
    }
  return r;
}

void OlContext::AddRule (ParseRule* rule)
{
  rules->push_back ((OlContext::Rule*) rule->GetRule ());
}

OlContext::Rule* OlContext::CreateAggRule (ParseTerm *lhs, ParseAggTerm *rhs,
  bool deleteFlag, ParseExpr *n)
{
  ParseFunctor *h = dynamic_cast<ParseFunctor*> (lhs);
  string ruleName = (n) ? n->value->ToString () : "";
  Rule *r = new Rule (ruleName, h, deleteFlag);
  r->terms.push_back (rhs);
  return r;
}

//
// Print out the whole parse result, if we can
//
string OlContext::ToString ()
{
  ostringstream r;

  // Errors first
  for (ErrorList::iterator e = errors.begin (); e != errors.end (); e++)
    {
      r << "error(" << (*e)->line_num << ",'" << (*e)->msg << "').\n";
    }
  for (TableInfoMap::iterator i = tables->begin (); i != tables->end (); i++)
    {
      r << i->second->ToString () << "\n";
    }

  for (RuleList::iterator rule = rules->begin (); rule != rules->end (); rule++)
    {
      r << (*rule)->ToString () << "\n";
    }
  return r.str ();
}

OlContext::OlContext () :
  lexer (NULL)
{
  rules = new RuleList ();
  tables = new TableInfoMap ();
  mContext.mId = NULL;
}

OlContext::~OlContext ()
{
  delete rules;
  delete tables;
}

void OlContext::ParseString (const char *prog)
{
  assert(lexer==NULL);
  lexer = new OlLexer (prog);
  ol_parser_parse (this);
  delete lexer;
  lexer = NULL;
}

void OlContext::ParseStream (istream *str)
{
  assert(lexer==NULL);
  lexer = new OlLexer (str);
  ol_parser_parse (this);
  delete lexer;
  lexer = NULL;
}

void OlContext::ReportError (string msg)
{
  NS_LOG_ERROR ("PARSER ERROR (line " << lexer->line_num () << "): " << msg);
  errors.push_back (new OlContext::Error (lexer->line_num (), msg));
}

void OlContext::Watch (ParseExpr *w, string modifiers)
{
  NS_LOG_INFO ("Added watched functor " << w->ToString ()
      << " with modifiers [" << modifiers << "]");
  watchTables.insert (make_pair (w->value->ToString (), modifiers));
}

void OlContext::Stage (ParseExpr *stageName, ParseExpr *inTupleName,
  ParseExpr *outTupleName)
{
  NS_LOG_INFO ("Add Stage Info " << stageName->ToString ()
      << " taking inTupleName= " << inTupleName->value->ToString ()
      << " outTupleName= " << outTupleName->value->ToString ());
  mStages[inTupleName->ToString ()] = new ExtStageSpec (stageName->ToString (),
    inTupleName->value->ToString (), outTupleName->value->ToString ());
}

void OlContext::TraceTuple (ParseExpr *w)
{
  NS_LOG_INFO ("Add trace variable " << w->ToString ());
  tuplesToTrace.insert (w->value->ToString ());
}

void OlContext::TraceTable (ParseExpr *w)
{
  NS_LOG_INFO ("Add traced table " << w->ToString ());
  tablesToTrace.insert (w->value->ToString ());
}

void OlContext::table (ParseExpr *name, ParseExpr *ttl, ParseExpr *size,
  ParseExprList *keys)
{
  TableInfo *tableInfo = new TableInfo ();
  tableInfo->tableName = name->ToString ();

  int myTtl =
      (dynamic_cast<ValInt32*> (PeekPointer (ttl->value)))->GetInt32Value ();
  if (myTtl == -1)
    {
      tableInfo->timeout = OlContext::NO_EXPIRATION;
    }
  else if (myTtl == 0)
    {
      ReportError ("bad timeout for materialized table");
    }
  else
    {
      tableInfo->timeout = myTtl;
    }

  int mySize =
      (dynamic_cast<ValInt32*> (PeekPointer (size->value)))->GetInt32Value ();
  // Hack because infinity token has a -1 value
  if (mySize == -1)
    {
      tableInfo->size = OlContext::NO_SIZE;
    }
  else
    {
      tableInfo->size = mySize;
    }

  if (keys)
    {
      for (ParseExprList::iterator i = keys->begin (); i != keys->end (); i++)
        {
          ParseKey* key = dynamic_cast<ParseKey*> (*i);
          ValInt32* val = dynamic_cast<ValInt32*> (PeekPointer (
            key->getAttribute ()->value));
          tableInfo->primaryKeys.push_back (val->GetInt32Value ());
          tableInfo->primaryKeyTypes.push_back (key->GetType ());
        }
    }
  else
    {
      // If keys was NULL, then I'm going to leave primaryKeys
      // alone, which means it is going to be empty
      // which means it's going to be KEYID
    }

  tables->insert (make_pair (tableInfo->tableName, tableInfo));

  NS_LOG_DEBUG ("Materialize " << tableInfo->tableName << "/" << ", timeout "
      << tableInfo->timeout << ", size " << tableInfo->size);
}

//
// Adding a fact [not used]
//
void OlContext::Fact (ParseTerm *term)
{
}

void OlContext::Query (ParseTerm *term)
{
  singleQuery = dynamic_cast<ParseFunctor*> (term);
  NS_LOG_INFO ("Add query " << singleQuery->ToString ());
}

bool OlContext::GotErrors ()
{
  return !errors.empty ();
}

void OlContext::DumpErrors ()
{
  for (ErrorList::iterator e = errors.begin (); e != errors.end (); e++)
    {
      NS_LOG_INFO ("error at line " << (*e)->line_num << ": '" << (*e)->msg
          << "'.");
    }
}

OlContext::ExtStageSpec::ExtStageSpec (string name, string input,
  string output) :
  stageName (name), inputTupleName (input), outputTupleName (output)
{
}

OlContext::ExtStageSpec::ExtStageSpec (const ExtStageSpec& s)
{
  stageName = s.stageName;
  inputTupleName = s.inputTupleName;
  outputTupleName = s.outputTupleName;
}

const OlContext::ExternalStageSpecMap&
OlContext::GetExtStagesInfo ()
{
  return mStages;
}
